import sys
import inspect
import importlib
import pkgutil
import inspect
from .base import GitSource

submodules = [ importlib.import_module(__name__ + '.' + name)
               for _,name,_ in pkgutil.walk_packages(sys.modules[__name__].__path__) ]

source_map = {} #map of source names to source class handlers.

for mod in submodules:
    for name,cls in inspect.getmembers(mod):
        if inspect.isclass(cls) and GitSource in cls.__bases__:
            source_map[name.lower()] = cls


from abc import ABC, abstractmethod

class GitSource(ABC):
    @abstractmethod
    def location(self, name):
        """Returns a URL corresponding to the location of repository {name}. Must
           be git clonable. Does not perform validation (confirm existence)"""
        pass

    @abstractmethod
    def create(self, name, is_private=True):
        """Creates a new repository called {name} within the given source."""
        pass

    @abstractmethod
    def list(self):
        """Returns all repositories contained within the provided source."""
        pass

    @abstractmethod
    def delete(self, name):
        """Deletes the named repository within the given sourc."""
        pass


import subprocess
from .base import GitSource

#TODO use paramiko + improve error handling

class SSH(GitSource):
    def __init__(self, name, host, user, project_dir):
        self.name = name
        self.host = host
        self.user = user
        self.project_dir = project_dir
        self.run = lambda cmd:\
            subprocess.check_output(["ssh", "{0}@{1}".format(user, host), cmd])

    def location(self, name):
        return "ssh://{user}@{host}:{path}".format(user=self.user,
                                                host=self.host,
                                                path=self.project_dir + "/" + name)
    
    def create(self, name, is_private=True):
        if not is_private:
            raise Exception("Cannot create public repository on standard SSH source.")
        
        cmd = "cd {projectDir} && mkdir {project} && cd {project} && git init --bare"\
               .format(projectDir=self.project_dir, project=name)
        try:
            self.run(cmd)
        except:
            raise Exception("Failed to create {0}, perhaps it already exists".format(name))
        
                                        
    def list(self):
        try:
            repos = self.run("ls " + self.project_dir).decode('utf-8')\
                       .strip()\
                       .split("\n")
            return [(repo, True) for repo in repos]
        except:
            raise Exception("Failed to list {0}, does the project directory exist?"
                            .format(self.project_dir))
        
    def delete(self, name):
        cmd = "cd {projectDir} && rm -r {name}".format(projectDir=self.project_dir, name=name)
        try:
            self.run(cmd)
        except:
            raise Exception("Failed to remove {0}, does it exist?".format(name))



import requests
import bs4
import re
from .base import GitSource

class Gitlab(GitSource):
    """
    Class for manipulating private projects hosted by gitlab (the service).
    Scraping is used in place of the API which is painfully slow
    (see https://gitlab.com/gitlab-com/support-forum/issues/576). The 
    constructor consumes an API access token which can be generated
    here: https://gitlab.com/profile/personal_access_tokens.
    """
    
    def __init__(self, name, api_token):
        self.name = name
        self.http = requests.Session()
        self.http.headers.update({'PRIVATE-TOKEN': api_token})
        self._user = None
        
    def user(self):
        if not self._user:
            r = re.search('username="([^"]*)', self.http.get('https://gitlab.com').text)
            if not r:
                raise Exception("Invalid private token supplied: " + token)
            self._user = r.group(1)
        
        return self._user
        
    def list(self):
        def _parse(content):
            root = bs4.BeautifulSoup(content, 'html.parser')
            nextpage = root.select_one('.pagination .next a')
            urls = []

            for project in root.findAll(class_="project-row"):
                url = project.find(attrs={'class': "project"})['href']
                private = project.find(class_="visibility-icon")['title'].startswith("Private")
                urls.append((url,private))

            return (urls,
                    'https://gitlab.com/' + nextpage['href'] if nextpage else None)
            
        results = set()
        cpage = 'http://gitlab.com/dashboard/projects'
        while cpage:
            r = self.http.get(cpage)
            if(r.status_code != 200):
                raise Exception("Project listing for {0} failed")
            (projects, cpage) = _parse(r.text)
            results.update(projects)
            
        return  { (re.sub("^/[^/]*/", "", r[0]), r[1]) for r in results }

    def delete(self, name):
        project_url = 'https://gitlab.com/' + self.user() + "/" + name
        r = self.http.get(project_url)
        if(r.status_code != 200):
            raise Exception("Project \"{0}\" does not seem to exist.".format(name))
        
        root = bs4.BeautifulSoup(r.text, 'html.parser')
        token = root.find(attrs={'name': 'csrf-token'})['content']

        r = self.http.post(project_url, {
            '_method': 'delete',
            'authenticity_token': token
        })
        if(r.status_code != 200):
            raise Exception("Failed to delete project, request output: \n\n" + r.text)


    def location(self, name):
        return "ssh://git@gitlab.com:/" + self.user() + "/" + name
    
    def create(self, name, description='', is_private=True):
        r = self.http.get('https://gitlab.com/projects/new')

        root = bs4.BeautifulSoup(r.text, 'html.parser')
        auth_token = root.find('meta', attrs={'name': 'csrf-token'})['content']
        users = root.find('select',
                          attrs={'name':
                                 'project[namespace_id]'})\
                    .find('optgroup',
                          attrs={'label': 'Users'})

        assert len(list(users.children)) == 1
        namespace_id = next(users.children)['value']

        params = {
            'utf8': '✓',
            'authenticity_token': auth_token,
            'project[namespace_id]': namespace_id,
            'project[path]': name,
            'project[description]': description,
            'project[visibility_level]': 0 if is_private else 20
        }
    
        r = self.http.post('https://gitlab.com/projects', params)
    
        if 'was successfully created.' not in r.text:
            raise Exception("Failed to create repo: {0}. Make sure it does not already exist."\
                            .format(name))



#!/usr/bin/env python3

from distutils.core import setup
setup(name='gt',
        version='0.3.02',
        author='Raheman Vaiya',
        author_email='r.vaiya@gmail.com',
        url='http://gitlab.com/rvaiya/gt',
        keywords='git github gitlab ssh cli console management',
        description='A simple tool for managing repositories located accross various git sources (e.g gitlab/github/unix boxes).',
        long_description=open('README.rst').read(),
        packages=['gt.sources'],
        scripts=['bin/gt'],
        classifiers=[
            'Programming Language :: Python :: 3',
            'Development Status :: 3 - Alpha'
            ],
      install_requires=['requests', 'bs4']
)
